import typing as T

import openai
from retry import retry

openai.api_key_path = "/home/vogel/data/openai-key.secret"

MAZE_STR = """
####X###
#### ###
# ##   #
#    # #
#### ###
##     #
## ### #
##.#####
"""

Maze = list[list[str]]
Dir = T.Literal["Up", "Down", "Left", "Right"]
Point = tuple[int, int]


def find_dot(maze: Maze) -> Point:
    y = next(i for i in range(len(maze)) if "." in maze[i])
    x = maze[y].index(".")
    return (x, y)


def move(maze: Maze, dir: Dir) -> T.Literal["impossible", "ok", "won"]:
    x, y = find_dot(maze)
    new_x, new_y = x, y
    match dir:
        case "Up":
            if y == 0:
                return "impossible"
            new_y = y - 1
        case "Down":
            if y == len(maze) - 1:
                return "impossible"
            new_y = y + 1
        case "Left":
            if x == 0:
                return "impossible"
            new_x = x - 1
        case "Right":
            if x == len(maze[0]) - 1:
                return "impossible"
            new_x = x + 1
    if maze[new_y][new_x] == " ":
        maze[y][x] = " "
        maze[new_y][new_x] = "."
        return "ok"
    elif maze[new_y][new_x] == "X":
        return "won"
    else:
        return "impossible"


def neighbors(maze: Maze) -> tuple[str, str, str, str]:
    x, y = find_dot(maze)
    left = "#" if x == 0 else maze[y][x - 1]
    right = "#" if x == len(maze[0]) - 1 else maze[y][x + 1]
    up = "#" if y == 0 else maze[y - 1][x]
    down = "#" if y == len(maze) - 1 else maze[y + 1][x]
    return (left, right, up, down)


SYSTEM_PROMPT = """
You are an expert maze-solving AI. You only respond with maze-related actions.
""".strip()

INITIAL_JS_PROMPT = """
You are in a maze. Your goal is to escape.
You can move in the four cardinal directions (north, south, east, west) using Javascript functions.
At each step, you'll be told what the squares at the cardinal directions are,
and you should reply with a move using a Javascript function call
(moveNorth(), moveSouth(), moveEast(), or moveWest()). Don't use any other functions or logic.
""".strip()

INITIAL_CMD_PROMPT = """
You are in a maze. Your goal is to escape.
You can move in the four cardinal directions (north, south, east, west) using special commands.
At each step, you'll be told what the squares at the cardinal directions are,
and you should reply with a move using a command
(MOVE {North, South, East, West}, e.g. MOVE NORTH or MOVE EAST). Don't use any other functions or logic.
""".strip()

INITIAL_PROMPT = INITIAL_CMD_PROMPT

NAMES = {
    "#": "Wall",
    " ": "Empty",
    "X": "Exit",
}

CONTEXT_PROMPT = """
Current position: west is {}, east is {}, north is {}, south is {}
""".strip()


def parse_direction(msg: str) -> Dir | None:
    if len(msg) > 15:
        return None  # probably repeating the prompt or something
    dirs: list[tuple[Dir, bool]] = [
        ("Up", "north" in msg.lower()),
        ("Down", "south" in msg.lower()),
        ("Left", "west" in msg.lower()),
        ("Right", "east" in msg.lower()),
    ]
    if [t[1] for t in dirs].count(True) != 1:
        return None
    return next(t[0] for t in dirs if t[1])

@retry(delay=1, backoff=2, max_delay=120)
def request_completion(messages) -> T.Any:
    return openai.ChatCompletion.create(
        model="gpt-4",
        temperature=0.7,
        messages=messages,
        request_timeout=20,
    )


def run(initial_prompt: str) -> tuple[int, list[Dir]]:
    maze = [list(line.strip()) for line in MAZE_STR.split("\n") if line.strip() != ""]

    print(initial_prompt)

    messages = [
        {"role": "system", "content": SYSTEM_PROMPT},
        {"role": "user", "content": initial_prompt},
    ]

    # this includes reprompts, which we want because better initial prompts should require
    # less reprompting
    steps = 0
    history = []

    while True:
        steps += 1
        if steps > 35:
            break  # doomed
        print("-" * 40)

        context = CONTEXT_PROMPT.format(*(NAMES[n] for n in neighbors(maze)))
        print(context)
        print()
        messages.append({"role": "user", "content": context})

        response = request_completion(messages)
        assert response["choices"][0]["finish_reason"] == "stop"

        newmsg = response["choices"][0]["message"]

        direction = parse_direction(newmsg["content"])
        print(f"AI: {newmsg['content']} (parsed as {direction})")
        history.append(direction)

        if direction is None:
            print("Reprompting...")
        else:
            messages.append({"role": "assistant", "content": newmsg["content"]})
            match move(maze, direction):
                case "impossible":
                    messages.append(
                        {"role": "user", "content": "You can't go that way."}
                    )
                case "ok":
                    print("\n".join("".join(row) for row in maze))
                case "won":
                    print("AI won!")
                    break
    return steps, history


if __name__ == "__main__":
    total_js_steps = 0
    js_hist = []
    for i in range(10):
        steps, history = run(INITIAL_JS_PROMPT)
        total_js_steps += steps
        js_hist.append((steps, history))
    avg_js_steps = total_js_steps / 10

    total_dsl_steps = 0
    dsl_hist = []
    for i in range(10):
        steps, history = run(INITIAL_CMD_PROMPT)
        total_dsl_steps += steps
        dsl_hist.append((steps, history))
    avg_dsl_steps = total_dsl_steps / 10

    print("-" * 40)
    print("JS history", js_hist)
    print("DSL history", dsl_hist)
    print(f"Average JS steps: {avg_js_steps:.2f}")
    print(f"Average DSL steps: {avg_dsl_steps:.2f}")
